@@ -28,7 +28,7 @@ BEGIN {
     }
 }
 
-our $VERSION = '0.31';
+our $VERSION = '0.32';
 
 sub _ANYPROCESS                    { return -1 }
 sub _COMMAND                       { return 0 }
@@ -155,14 +155,82 @@ sub _xvfb_binary {
     return 'Xvfb';
 }
 
+sub _dev_fd_works {
+    my ($self) = @_;
+    my $test_handle =
+      File::Temp::tempfile( File::Spec->tmpdir(),
+        'firefox_marionette_dev_fd_test_XXXXXXXXXXX' )
+      or Carp::Croak("Failed to open temporary file:$EXTENDED_OS_ERROR");
+    my @stats = stat '/dev/fd/' . fileno $test_handle;
+    if ( scalar @stats ) {
+        return 1;
+    }
+    elsif ( $OSNAME eq 'freebsd' ) {
+        Carp::carp(
+q[/dev/fd is not working.  Perhaps you need to mount fdescfs like so 'sudo mount -t fdescfs fdesc /dev/fd']
+        );
+    }
+    else {
+        Carp::carp("/dev/fd is not working for $OSNAME");
+    }
+    return 0;
+}
+
+sub _dbus_works {
+    my ($self)   = @_;
+    my $binary   = 'dbus-launch';
+    my $dev_null = File::Spec->devnull();
+    if ( my $pid = fork ) {
+        waitpid $pid, 0;
+        if ( $CHILD_ERROR == 0 ) {
+            return 1;
+        }
+        elsif ( $OSNAME eq 'freebsd' ) {
+            my @stats = stat '/etc/machine-id';
+            if ( scalar @stats ) {
+            }
+            else {
+                Carp::carp(
+q[D-Bus is not working.  Perhaps you need to create '/etc/machine-id' like so 'sudo dbus-uuidgen --ensure=/etc/machine-id']
+                );
+            }
+        }
+    }
+    elsif ( defined $pid ) {
+        eval {
+            if ( !$self->_debug() ) {
+                open STDOUT, q[>], $dev_null
+                  or Carp::croak(
+                    "Failed to redirect STDOUT to $dev_null:$EXTENDED_OS_ERROR"
+                  );
+                open STDERR, q[>], $dev_null
+                  or Carp::croak(
+                    "Failed to redirect STDERR to $dev_null:$EXTENDED_OS_ERROR"
+                  );
+            }
+            exec {$binary} $binary
+              or Carp::croak("Failed to exec '$binary':$EXTENDED_OS_ERROR");
+        } or do {
+            chomp $EVAL_ERROR;
+            warn "$EVAL_ERROR\n";
+        };
+        exit 1;
+    }
+    return 0;
+}
+
 sub _xvfb_exists {
     my ($self)   = @_;
     my $binary   = $self->_xvfb_binary();
     my $dev_null = File::Spec->devnull();
-    eval {
-        require Digest::MD5;
-        require Crypt::URandom;
-    } or do {
+    if ( !$self->_dev_fd_works() ) {
+        return 0;
+    }
+    if ( !$self->_dbus_works() ) {
+        return 0;
+    }
+    eval { require Crypt::URandom; } or do {
+        Carp::croak('Unable to load Crypt::URandom');
         return 0;
     };
     if ( my $pid = fork ) {
@@ -199,9 +267,8 @@ sub xvfb {
 
 sub _launch_xauth {
     my ( $self, $display_number ) = @_;
-    my $mcookie =
-      Digest::MD5::md5_hex(
-        Crypt::URandom::urandom( _NUMBER_OF_MCOOKIE_BYTES() ) );
+    my $mcookie = unpack 'H*',
+      Crypt::URandom::urandom( _NUMBER_OF_MCOOKIE_BYTES() );
     my $source_handle =
       File::Temp::tempfile( File::Spec->tmpdir(),
         'firefox_marionette_xauth_source_XXXXXXXXXXX' )
@@ -487,7 +554,7 @@ sub _reap {
     }
     else {
         while ( ( my $pid = waitpid _ANYPROCESS(), POSIX::WNOHANG() ) > 0 ) {
-            if ( $pid == $self->_pid() ) {
+            if ( ( $self->_pid() ) && ( $pid == $self->_pid() ) ) {
                 $self->{_child_error} = $CHILD_ERROR;
             }
             elsif ( ( $self->xvfb() ) && ( $pid == $self->xvfb() ) ) {
@@ -620,7 +687,7 @@ sub new_session {
     if (   ( defined $capabilities )
         && ( $capabilities->isa('Firefox::Marionette::Capabilities') ) )
     {
-        $parameters = {
+        my $actual = {
             acceptInsecureCerts => $capabilities->accept_insecure_certs()
             ? JSON::true()
             : JSON::false(),
@@ -632,11 +699,34 @@ sub new_session {
               $capabilities->moz_accessibility_checks() ? JSON::true()
             : JSON::false(),
         };
+        $parameters = $actual;    # for Mozilla 57 and after
+        foreach my $key ( sort { $a cmp $b } keys %{$actual} ) {
+            $parameters->{capabilities}->{requiredCapabilities}->{$key} =
+              $actual->{$key};    # for Mozilla 56 (and below???)
+        }
     }
     my $message_id = $self->_new_message_id();
     $self->_send_request(
         [ _COMMAND(), $message_id, 'WebDriver:NewSession', $parameters ] );
-    my $response = $self->_get_response($message_id);
+    my $response;
+    eval { $response = $self->_get_response($message_id); } or do {
+        if (
+            ( ref $EVAL_ERROR )
+            && ( ( ref $EVAL_ERROR ) eq
+                'Firefox::Marionette::Exception::Response' )
+            && ( $EVAL_ERROR->error() eq 'invalid session id' )
+          )
+        {
+            my $fallback_message_id = $self->_new_message_id();
+            $self->_send_request(
+                [ _COMMAND(), $fallback_message_id, 'newSession', $parameters ]
+            );
+            $response = $self->_get_response($fallback_message_id);
+        }
+        else {
+            Carp::croak($EVAL_ERROR);
+        }
+    };
     $self->{session_id} = $response->result()->{sessionId};
     return ( $self->{session_id},
         $self->_create_capabilities( $response->result()->{capabilities} ) );
@@ -1382,14 +1472,20 @@ sub quit {
         }
     }
     elsif ( $self->_socket() ) {
-        $flags ||=
-          ['eAttemptQuit'];    # ["eConsiderQuit", "eAttemptQuit", "eForceQuit"]
-        my $message_id = $self->_new_message_id();
-        $self->_send_request(
-            [ _COMMAND(), $message_id, 'Marionette:Quit', { flags => $flags } ]
-        );
-        my $response = $self->_get_response($message_id);
-        my $socket   = delete $self->{_socket};
+        if ( $self->_session_id() ) {
+            $flags ||=
+              ['eAttemptQuit']
+              ;    # ["eConsiderQuit", "eAttemptQuit", "eForceQuit"]
+            my $message_id = $self->_new_message_id();
+            $self->_send_request(
+                [
+                    _COMMAND(), $message_id,
+                    'Marionette:Quit', { flags => $flags }
+                ]
+            );
+            my $response = $self->_get_response($message_id);
+        }
+        my $socket = delete $self->{_socket};
         close $socket
           or
           Carp::croak("Failed to close socket to firefox:$EXTENDED_OS_ERROR");
@@ -1399,6 +1495,7 @@ sub quit {
         }
         else {
             while ( kill 0, $self->_pid() ) {
+                sleep 1;
                 $self->_reap();
             }
         }
@@ -1413,7 +1510,6 @@ sub quit {
         }
         delete $self->{_xvfb_display_number};
         delete $self->{_xvfb_authority_directory};
-        delete $self->{_xvfb_pid};
     }
     return $self->child_error();
 }
@@ -1432,7 +1528,7 @@ sub _terminate_process {
             }
         }
     }
-    elsif ( kill 0, $self->_pid() ) {
+    elsif ( ( $self->_pid() ) && ( kill 0, $self->_pid() ) ) {
         my $term_signal = $self->_signal_number('TERM')
           ;    # https://support.mozilla.org/en-US/questions/752748
         if ( $term_signal > 0 ) {
@@ -1691,7 +1787,15 @@ sub _send_request {
     }
     my $result = syswrite $self->_socket(), "$length:$json";
     if ( !defined $result ) {
-        Carp::croak("Failed to send request to firefox:$EXTENDED_OS_ERROR");
+        my $socket_error = $EXTENDED_OS_ERROR;
+        if ( $self->alive() ) {
+            Carp::croak("Failed to send request to firefox:$socket_error");
+        }
+        else {
+            my $error_message =
+              $self->error_message() ? $self->error_message() : q[];
+            Carp::croak($error_message);
+        }
     }
 }
 
@@ -1704,6 +1808,19 @@ sub _read_from_socket {
         if ( defined $number_of_bytes ) {
             $initial_buffer .= $octet;
         }
+        else {
+            my $socket_error = $EXTENDED_OS_ERROR;
+            if ( $self->alive() ) {
+                Carp::croak(
+"Failed to read size of response from socket to firefox:$socket_error"
+                );
+            }
+            else {
+                my $error_message =
+                  $self->error_message() ? $self->error_message() : q[];
+                Carp::croak($error_message);
+            }
+        }
         if ( $initial_buffer =~ s/^(\d+)://smx ) {
             ($number_of_bytes_in_response) = ($1);
         }
@@ -1721,8 +1838,17 @@ sub _read_from_socket {
             $number_of_bytes_already_read += $number_of_bytes_read;
         }
         else {
-            Carp::croak(
-                "Failed to read from socket to firefox:$EXTENDED_OS_ERROR");
+            my $socket_error = $EXTENDED_OS_ERROR;
+            if ( $self->alive() ) {
+                Carp::croak(
+"Failed to read response from socket to firefox:$socket_error"
+                );
+            }
+            else {
+                my $error_message =
+                  $self->error_message() ? $self->error_message() : q[];
+                Carp::croak($error_message);
+            }
         }
     }
     if ( $self->_debug() ) {
@@ -1776,7 +1902,7 @@ Firefox::Marionette - Automate the Firefox browser with the Marionette protocol
 
 =head1 VERSION
 
-Version 0.31
+Version 0.32
 
 =head1 SYNOPSIS
 
