@@ -4,7 +4,7 @@ use strict;
 use warnings;
 use overload '""' => 'string';
 
-our $VERSION = '0.31';
+our $VERSION = '0.32';
 
 sub throw {
     my ( $class, $response ) = @_;
@@ -12,6 +12,11 @@ sub throw {
     Carp::croak($self);
 }
 
+sub error {
+    my ($self) = @_;
+    return $self->{response}->error()->{error};
+}
+
 sub string {
     my ($self) = @_;
     return $self->{response}->error()->{error} . q[: ]
@@ -32,7 +37,7 @@ Firefox::Marionette::Exception::Response - Represents a Marionette protocol resp
 
 =head1 VERSION
 
-Version 0.31
+Version 0.32
 
 =head1 SYNOPSIS
 
@@ -53,6 +58,10 @@ accepts a Marionette L<response|Firefox::Marionette::Response> as it's only para
 
 returns a stringified version of the error and message fields on the error.
 
+=head2 error
+
+returns the firefox error message
+
 =head2 trace
 
 returns the firefox trace.
